/*
	Custom Emulator Firmware
	Copyright (C) 2012, Total_Noob

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <pspsdk.h>
#include <pspkernel.h>
#include <psploadexec_kernel.h>
#include <pspthreadman_kernel.h>

#include <stdio.h>
#include <string.h>

#include "main.h"
#include "../rebootex/rebootex.h"

int sctrlHENIsSE()
{
	return 1;
}

int sctrlHENIsDevhook()
{
	return 0;
}

int sctrlHENGetVersion()
{
	return 0x00001020;
}

int sctrlSEGetVersion()
{
	return 0x00020005;
}

PspIoDrv *sctrlHENFindDriver(char *drvname)
{
	int k1 = pspSdkSetK1(0);

	SceModule2 *mod = sceKernelFindModuleByName660("sceIOFileManager");
	u32 text_addr = mod->text_addr;

	u32 *(* GetDevice)(char *) = NULL;

	int i;
	for(i = 0; i < (int)mod->text_size; i += 4)
	{
		u32 addr = text_addr + i;
		if(_lw(addr) == 0xA2200000)
		{
			GetDevice = (void *)K_EXTRACT_CALL(addr + 4);
			break;
		}
	}

	if(!GetDevice)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	u32 *u = GetDevice(drvname);
	if(!u)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	pspSdkSetK1(k1);

	return (PspIoDrv *)u[1];
}

typedef struct
{
	u32 magic;
	u32 size;
	u32 path_len;
} TNPackage;

typedef struct
{
	char *name;
	void *buffer;
	u32 size;
} FileList;

extern RebootexConfig rebootex_config;

int MakeFileList()
{
	/* Open package */
	char path[64];
	strcpy(path, rebootex_config.savedata_path);
	strcat(path, "/FLASH0.TN");
	SceUID fd = sceIoOpen(path, PSP_O_RDONLY, 0);

	if (fd < 0)
		return fd;

	/* Get file_n */
	u32 file_n = 0;

	int size = sceIoLseek(fd, 0, PSP_SEEK_END);
	sceIoLseek(fd, size - 4, PSP_SEEK_SET);
	sceIoRead(fd, &file_n, sizeof(file_n));
	sceIoLseek(fd, 0, PSP_SEEK_SET);

	/* Get file_list size */
	FileList *file_list = (FileList *)0x8B000000;
	while(file_list->buffer != 0) file_list++;

	u32 file_list_size = (u32)file_list - 0x8B000000;

	/* Add new files */
	memcpy((void *)0x8BA00000 + file_n * sizeof(FileList), (void *)0x8B000000, file_list_size);
	u32 pointer = 0x8BA00000 + file_list_size + (file_n + 1) * sizeof(FileList);
	file_list = (FileList *)0x8BA00000;

	u32 path_pointer = 0x8B000000;

	int res = 0;
	do
	{
		TNPackage package;
		res = sceIoRead(fd, &package, sizeof(TNPackage));

		if (package.magic == 0x4B504E54) {
			/* Add path */
			file_list->name = (void *)path_pointer;
			sceIoRead(fd, file_list->name, package.path_len);
			path_pointer += package.path_len;

			/* Add buffer */
			pointer = (((u32)pointer + 0x40) & ~0x3F);
			file_list->buffer = (void *)pointer;
			sceIoRead(fd, (void *)file_list->buffer, package.size);
			pointer += package.size;

			/* Add size */
			file_list->size = package.size;

			/* Next entry */
			file_list++;
		}
	} while(res > 0);

	sceIoClose(fd);

	return res;
}

int (* flashLoad)(int);

int flashLoadPatch(int cmd)
{
	int ret = flashLoad(cmd);

	if ((cmd == 1058) && (ret != -1))
		MakeFileList();

	return ret;
}

int sctrlKernelLoadExecVSHWithApitype(int apitype, const char *file, struct SceKernelLoadExecVSHParam *param)
{
	int k1 = pspSdkSetK1(0);

	SceModule2 *mod = sceKernelFindModuleByName660("sceLoadExec");
	u32 text_addr = mod->text_addr;

	SceModule2 *mod2 = sceKernelFindModuleByName660("sceKermitPeripheral_Driver");

	flashLoad = (void *)(mod2->text_addr + 0x88C);

	_sw((u32)&flashLoadPatch, text_addr + 0x66A0);

	ClearCaches();

	int (* LoadExecVSH)(int apitype, const char *file, struct SceKernelLoadExecVSHParam *param, int unk2) = (void *)text_addr + 0x23D0;

	int res = LoadExecVSH(apitype, file, param, 0x10000);
	pspSdkSetK1(k1);
	return res;
}

int sctrlKernelExitVSH(struct SceKernelLoadExecVSHParam *param)
{
	int k1 = pspSdkSetK1(0);
	int res = sceKernelExitVSHVSH660(param);
	pspSdkSetK1(k1);
	return res;
}

int sctrlKernelLoadExecVSHMs1(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_MS1, file, param);
}

int sctrlKernelLoadExecVSHMs2(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_MS2, file, param);
}

int sctrlKernelLoadExecVSHMs3(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_MS3, file, param);
}

int sctrlKernelLoadExecVSHMs4(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_MS4, file, param);
}

int sctrlKernelLoadExecVSHDisc(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_DISC, file, param);
}

int sctrlKernelLoadExecVSHDiscUpdater(const char *file, struct SceKernelLoadExecVSHParam *param)
{
	return sctrlKernelLoadExecVSHWithApitype(PSP_INIT_APITYPE_DISC_UPDATER, file, param);
}

int sctrlKernelSetUserLevel(int level)
{
	int k1 = pspSdkSetK1(0);
	int res = sceKernelGetUserLevel();

	SceModule2 *mod = sceKernelFindModuleByName660("sceThreadManager");
	u32 text_addr = mod->text_addr;

	u32 *thstruct = (u32 *)_lw(text_addr + 0x19F00);
	thstruct[0x14/4] = (level ^ 8) << 28;

	pspSdkSetK1(k1);
	return res;
}